//
// file: primitive3D_frustumLight.fx
// version: v1.5
// author: Ren712
//

//--------------------------------------------------------------------------------------
// Settings
//--------------------------------------------------------------------------------------
float3 sElementPosition = float3(0, 0, 0);
float3 sElementRotation = float3(0, 0, 0);
float4 sElementColor = float4(0, 0, 0, 0);
float3 sElementSize = float3(1,1,1);
float3 sClipOffset = float3(0,0,0.2);
float3 sFogVector = float3(0,0,1);

float sLightPhiX = 0.2; // Phi is the outer cone angle
float sLightPhiY = 0.5; // Phi is the outer cone angle
float sLightAttenuation = 20; // light attenuation (max radius)
float sLightAttenuationPower = 2; // light attenuation power

float sCoronaRescale = 1;
float sCoronaDepthBias = 0.2;
float sCoronaScaleSpread = 0.85;

bool sFogEnable = false;

bool fZEnable = true;
bool fZWriteEnable = true;

float fDepthBias = 1;
float2 gDistFade = float2(250, 150);

//--------------------------------------------------------------------------------------
// Variables set by MTA
//--------------------------------------------------------------------------------------
float4x4 gProjection : PROJECTION;
float4x4 gView : VIEW;
float4x4 gViewInverse : VIEWINVERSE;
float3 gCameraPosition : CAMERAPOSITION;
static const float PI = 3.14159265f;
int CUSTOMFLAGS < string skipUnusedParameters = "yes"; >;

//--------------------------------------------------------------------------------------
// Structures
//--------------------------------------------------------------------------------------
struct VSInput
{
    float3 Position : POSITION0;
    float2 TexCoord : TEXCOORD0;
    float4 Diffuse : COLOR0;
};

struct PSInput
{
    float4 Position : POSITION0;
    float2 TexCoord : TEXCOORD0;
    float4 Diffuse : COLOR0;
    float DistFade : TEXCOORD1;
    float DistFactor : TEXCOORD2;
    float PlaneDist : TEXCOORD3;
};

//--------------------------------------------------------------------------------------
// Returns a rotation matrix (rotate by X)
//--------------------------------------------------------------------------------------
float4x4 makeXRotation( float angleInRadians) 
{
  float c = cos(angleInRadians);
  float s = sin(angleInRadians);

  return float4x4(
    1, 0, 0, 0,
    0, c, s, 0,
    0, -s, c, 0,
    0, 0, 0, 1
  );
};

//--------------------------------------------------------------------------------------
// Returns a rotation matrix (rotate by Y)
//--------------------------------------------------------------------------------------
float4x4 makeYRotation( float angleInRadians) 
{
  float c = cos(angleInRadians);
  float s = sin(angleInRadians);

  return float4x4(
    c, 0, -s, 0,
    0, 1, 0, 0,
    s, 0, c, 0,
    0, 0, 0, 1
  );
};

//--------------------------------------------------------------------------------------
// Returns a rotation matrix (rotate by Z)
//--------------------------------------------------------------------------------------
float4x4 makeZRotation( float angleInRadians) 
{
  float c = cos(angleInRadians);
  float s = sin(angleInRadians);
  
  return float4x4(
     c, s, 0, 0,
    -s, c, 0, 0,
     0, 0, 1, 0,
     0, 0, 0, 1
  );
}

//--------------------------------------------------------------------------------------
// Create world matrix with world position and euler rotation
//--------------------------------------------------------------------------------------
float4x4 createWorldMatrix(float3 pos, float3 rot)
{
    float4x4 eleMatrix = {
        float4(cos(rot.z) * cos(rot.y) - sin(rot.z) * sin(rot.x) * sin(rot.y), 
                cos(rot.y) * sin(rot.z) + cos(rot.z) * sin(rot.x) * sin(rot.y), -cos(rot.x) * sin(rot.y), 0),
        float4(-cos(rot.x) * sin(rot.z), cos(rot.z) * cos(rot.x), sin(rot.x), 0),
        float4(cos(rot.z) * sin(rot.y) + cos(rot.y) * sin(rot.z) * sin(rot.x), sin(rot.z) * sin(rot.y) - 
                cos(rot.z) * cos(rot.y) * sin(rot.x), cos(rot.x) * cos(rot.y), 0),
        float4(pos.x,pos.y,pos.z, 1),
    };
    return eleMatrix;
}

//--------------------------------------------------------------------------------------
// Vertex Shader 
//--------------------------------------------------------------------------------------
PSInput VertexShaderFunction(VSInput VS, uniform float4x4 rotMat, uniform float3 offset)
{	
    PSInput PS = (PSInput)0;

    // set proper position and scale of the quad
    VS.Position.xyz = float3(- 0.5 + VS.TexCoord.xy, 0);
	
    VS.TexCoord.y = 1 - VS.TexCoord.y;	

    VS.Position.xyz = mul(VS.Position.xyz, (float3x3)rotMat);
    VS.Position.xyz += offset;
    VS.Position.xyz += sClipOffset;	
	
    if (VS.Position.x > 0) VS.Position.xyz += float3(1,0,0) * VS.Position.z;
        else VS.Position.xyz -= float3(1,0,0) * VS.Position.z;

    if (VS.Position.y > 0) VS.Position.xyz += float3(0,1,0) * VS.Position.z; 
        else VS.Position.xyz -= float3(0,1,0) * VS.Position.z;

    VS.Position.xyz -= sClipOffset;
    VS.Position.z += 0.5;
	
    PS.PlaneDist = dot(sFogVector, VS.Position.xyz);
	
    // get size
    float3 elementSize = float3(2 * tan(sLightPhiX) * sLightAttenuation, 2 * tan(sLightPhiY) * sLightAttenuation, sLightAttenuation);
	
    VS.Position.xyz *= elementSize.yxz;
    VS.Position.xyz = VS.Position.xzy;

    VS.TexCoord.x = 1 - VS.TexCoord.x;

    // create WorldMatrix for the quad
    float4x4 sWorld = createWorldMatrix(sElementPosition, sElementRotation);

    // get element's view direction and distance from camera
    float3 viewDirection = (gCameraPosition - sElementPosition);
	
    // interpolate between shafts and corona
    float3 altPosXf = sWorld[2].xyz * dot(sWorld[2].xyz, viewDirection);
    float3 conePosXf = sWorld[1].xyz * sClipOffset * 2 * tan(sLightPhiY);
    float angleXf = acos(dot(normalize((viewDirection - altPosXf) - conePosXf), sWorld[1].xyz)); 
    float impactX = pow( smoothstep(sLightPhiY * 2.5, sLightPhiY * 0.5, angleXf ), 1);
    if (angleXf > sLightPhiY * 2.5) impactX = 0;
    float3 altPosYf = sWorld[0].xyz * dot(sWorld[0].xyz, viewDirection);
    float3 conePosYf = sWorld[1].xyz * sClipOffset * 2 * tan(sLightPhiX);
    float angleYf = acos(dot(normalize((viewDirection - altPosYf) - conePosYf), sWorld[1].xyz)); 
    float impactY = pow( smoothstep(sLightPhiX * 2.5, sLightPhiX * 0.5, angleYf ), 1);
    if (angleYf > sLightPhiX * 2.5) impactX = 0;	
    float impactFull = saturate(min(impactX, impactY));

    // calculate screen position of the vertex
    float4 wPos = mul(float4(VS.Position, 1), sWorld);
    float4 vPos = mul(wPos, gView); 
    PS.Position = mul(vPos, gProjection);

    // get clip values
    float nearClip = - gProjection[3][2] / gProjection[2][2];
    float farClip = (gProjection[3][2] / (1 - gProjection[2][2]));	
	
    // fade corona
    float DistFromCam = distance(gCameraPosition, sElementPosition.xyz);
    float medianSize = max(max(elementSize.x, elementSize.y), elementSize.z) /3;
    float2 DistFade = float2(min(gDistFade.x, farClip - medianSize / 2), min(gDistFade.y, farClip - medianSize /2 - (gDistFade.x - gDistFade.y)));
    PS.DistFade = saturate((DistFromCam - DistFade.x)/(DistFade.y - DistFade.x));
	
    DistFromCam /= fDepthBias;
    PS.DistFactor = saturate(DistFromCam * 0.5 - 1.6);	

    // pass texCoords and vertex color to PS
    PS.TexCoord = VS.TexCoord;
    PS.Diffuse = VS.Diffuse * sElementColor;
	
    PS.Diffuse.a *= saturate(1.5 - impactFull);
	
    return PS;
}

PSInput VertexShaderFunction2(VSInput VS)
{
    PSInput PS = (PSInput)0;
	
    // set proper position and scale of the quad
    VS.Position.xyz = float3(- 0.5 + VS.TexCoord.xy, 0);
	
    VS.TexCoord.y = 1 - VS.TexCoord.y;
	
    VS.Position.xyz += sClipOffset;
    VS.Position.z -= 0.5;
	
    if (VS.Position.x > 0) VS.Position.xyz += float3(1,0,0) * VS.Position.z;
        else VS.Position.xyz -= float3(1,0,0) * VS.Position.z;

    if (VS.Position.y > 0) VS.Position.xyz += float3(0,1,0) * VS.Position.z; 
        else VS.Position.xyz -= float3(0,1,0) * VS.Position.z;	
		
    VS.Position.z += 0.5;
    VS.Position.xyz -= sClipOffset;
	
    // get size
    float3 elementSize = float3(2 * tan(sLightPhiX) * sLightAttenuation, 2 * tan(sLightPhiY) * sLightAttenuation, sLightAttenuation);
	
    VS.Position.xyz *= elementSize.yxz * sCoronaRescale;
    VS.Position.xyz = VS.Position.xzy;
	
    VS.TexCoord.x = 1 - VS.TexCoord.x;

    // create WorldMatrix for the quad
    float4x4 sWorld = createWorldMatrix(sElementPosition, sElementRotation);

    // get element's view direction and distance from camera
    float3 viewDirection = (gCameraPosition - sElementPosition);
    float viewDistance = length(viewDirection);
	
    // is camera at the left/right side of the plane
    float planeYDist = dot(sWorld[1].xyz, viewDirection);
    float angYDir = planeYDist < 0 ?  1 : -1;
	
    // get angle between left plane and view
    float fwDist = dot(sWorld[0].xyz, viewDirection);
    float3 sLinePosition = sElementPosition + sWorld[0].xyz * fwDist;
    float angleY = acos(dot(normalize(sLinePosition - gCameraPosition), sWorld[2].xyz));
	
    // interpolate between shafts and corona
    float3 altPosXf = sWorld[2].xyz * dot(sWorld[2].xyz, viewDirection);
    float3 conePosXf = sWorld[1].xyz * sClipOffset * 2 * tan(sLightPhiY);
    float angleXf = acos(dot(normalize((viewDirection - altPosXf) - conePosXf), sWorld[1].xyz)); 
    float impactX = pow( smoothstep(sLightPhiY * 2.5, sLightPhiY * 0.5, angleXf ), 1);
    if (angleXf > sLightPhiY * 2) impactX = 0;
    float3 altPosYf = sWorld[0].xyz * dot(sWorld[0].xyz, viewDirection);
    float3 conePosYf = sWorld[1].xyz * sClipOffset * 2 * tan(sLightPhiX);
    float angleYf = acos(dot(normalize((viewDirection - altPosYf) - conePosYf), sWorld[1].xyz)); 
    float impactY = pow( smoothstep(sLightPhiX * 2.5, sLightPhiX * 0.5, angleYf ), 1);
    if (angleYf > sLightPhiX * 2.5) impactX = 0;	
    float impactFull = saturate(min(impactX, impactY));

    // scale the quad based on impactFull
    VS.Position.xz *= 1.5 + impactFull * sCoronaScaleSpread * 5;
    float3 fwVec = sWorld[1].xyz;
	
    // rotate the material matrix to recreate cylindrical billboard effect
    sWorld = mul(makeXRotation(angYDir * (- angleY - PI / 2)), sWorld);
	
    // move the billboard a bit forward
    sWorld[3].xyz += fwVec * sCoronaDepthBias;

    // calculate screen position of the vertex
    float4 worldPos = mul(float4(VS.Position, 1), sWorld);
    float4 vPos = mul(worldPos, gView); 
    vPos.xyz += fDepthBias * 1.5 * mul( normalize(gCameraPosition - sElementPosition), (float3x3)gView).xyz;
    PS.Position = mul(vPos, gProjection);
	
    // get clip values
    float nearClip = - gProjection[3][2] / gProjection[2][2];
    float farClip = (gProjection[3][2] / (1 - gProjection[2][2]));	
	
    // fade corona
    float DistFromCam = distance(gCameraPosition, sElementPosition.xyz);
    float medianSize = max(max(elementSize.x, elementSize.y), elementSize.z);
    float2 DistFade = float2(min(gDistFade.x, farClip - medianSize / 2), min(gDistFade.y, farClip - medianSize /2 - (gDistFade.x - gDistFade.y)));
    PS.DistFade = saturate((DistFromCam - DistFade.x)/(DistFade.y - DistFade.x));
	
    DistFromCam /= fDepthBias;
    PS.DistFactor = saturate( DistFromCam * 0.5 - 1.6 );	

    // pass texCoords and vertex color to PS
    PS.TexCoord = VS.TexCoord;
    PS.Diffuse = VS.Diffuse * sElementColor;
	
    return PS;
}

//--------------------------------------------------------------------------------------
// Pixel shaders 
//--------------------------------------------------------------------------------------
float4 PixelShaderFunction(PSInput PS) : COLOR0
{
    float fogFactor = 1 - saturate((PS.PlaneDist - 0.5)/ (1 - 0.5));
	
    // apply diffuse color
    float4 finalColor = PS.Diffuse;
    finalColor.a *= fogFactor;

    return saturate(finalColor);
}

float4 PixelShaderFunction2(PSInput PS) : COLOR0
{
    // compute the distance attenuation factor
    float fDistance = length(PS.TexCoord.xy - 0.5);

    // compute the attenuation
    float fAttenuation = 1 - saturate(pow(fDistance * 2,2));
	
    // apply diffuse color
    float4 finalColor = PS.Diffuse;
	
    finalColor *= fAttenuation;
    finalColor.a *= PS.DistFactor;
    finalColor.a *= saturate(PS.DistFade);

    return saturate(finalColor);
}

//--------------------------------------------------------------------------------------
// Techniques
//--------------------------------------------------------------------------------------
technique dxDrawPrimitive3Dfx_frustumLight
{
  pass P0
  {
    ZEnable = fZEnable;
    ZFunc = LessEqual;
    ZWriteEnable = fZWriteEnable;
    CullMode = 1;
    ShadeMode = Gouraud;
    AlphaBlendEnable = true;
    SrcBlend = SrcAlpha;
    DestBlend = One;
    AlphaTestEnable = true;
    AlphaRef = 1;
    AlphaFunc = GreaterEqual;
    Lighting = false;
    FogEnable = sFogEnable;
    VertexShader = compile vs_2_0 VertexShaderFunction(makeZRotation(0), float3(0,0,-0.5));
    PixelShader  = compile ps_2_0 PixelShaderFunction();
  }
  
  pass P1
  {
    ZEnable = fZEnable;
    ZFunc = LessEqual;
    ZWriteEnable = fZWriteEnable;
    CullMode = 1;
    ShadeMode = Gouraud;
    AlphaBlendEnable = true;
    SrcBlend = SrcAlpha;
    DestBlend = One;
    AlphaTestEnable = true;
    AlphaRef = 1;
    AlphaFunc = GreaterEqual;
    Lighting = false;
    FogEnable = sFogEnable;
    VertexShader = compile vs_2_0 VertexShaderFunction(makeZRotation(PI), float3(0,0,0.5));
    PixelShader  = compile ps_2_0 PixelShaderFunction();
  }
  
  pass P2
  {
    ZEnable = fZEnable;
    ZFunc = LessEqual;
    ZWriteEnable = fZWriteEnable;
    CullMode = 1;
    ShadeMode = Gouraud;
    AlphaBlendEnable = true;
    SrcBlend = SrcAlpha;
    DestBlend = One;
    AlphaTestEnable = true;
    AlphaRef = 1;
    AlphaFunc = GreaterEqual;
    Lighting = false;
    FogEnable = sFogEnable;
    VertexShader = compile vs_2_0 VertexShaderFunction(makeYRotation(PI/2), float3(0.5,0,0));
    PixelShader  = compile ps_2_0 PixelShaderFunction();
  }
  
  pass P3
  {
    ZEnable = fZEnable;
    ZFunc = LessEqual;
    ZWriteEnable = fZWriteEnable;
    CullMode = 1;
    ShadeMode = Gouraud;
    AlphaBlendEnable = true;
    SrcBlend = SrcAlpha;
    DestBlend = One;
    AlphaTestEnable = true;
    AlphaRef = 1;
    AlphaFunc = GreaterEqual;
    Lighting = false;
    FogEnable = sFogEnable;
    VertexShader = compile vs_2_0 VertexShaderFunction(makeYRotation(-PI/2), float3(-0.5,0,0));
    PixelShader  = compile ps_2_0 PixelShaderFunction();
  }
  
  pass P4
  {
    ZEnable = fZEnable;
    ZFunc = LessEqual;
    ZWriteEnable = fZWriteEnable;
    CullMode = 1;
    ShadeMode = Gouraud;
    AlphaBlendEnable = true;
    SrcBlend = SrcAlpha;
    DestBlend = One;
    AlphaTestEnable = true;
    AlphaRef = 1;
    AlphaFunc = GreaterEqual;
    Lighting = false;
    FogEnable = sFogEnable;
    VertexShader = compile vs_2_0 VertexShaderFunction(makeXRotation(PI/2), float3(0,-0.5,0));
    PixelShader  = compile ps_2_0 PixelShaderFunction();
  }
  
  pass P5
  {
    ZEnable = fZEnable;
    ZFunc = LessEqual;
    ZWriteEnable = fZWriteEnable;
    CullMode = 1;
    ShadeMode = Gouraud;
    AlphaBlendEnable = true;
    SrcBlend = SrcAlpha;
    DestBlend = One;
    AlphaTestEnable = true;
    AlphaRef = 1;
    AlphaFunc = GreaterEqual;
    Lighting = false;
    FogEnable = sFogEnable;
    VertexShader = compile vs_2_0 VertexShaderFunction(makeXRotation(-PI/2), float3(0,0.5,0));
    PixelShader  = compile ps_2_0 PixelShaderFunction();
  }
  
  pass P6
  {
    ZEnable = fZEnable;
    ZFunc = LessEqual;
    ZWriteEnable = fZWriteEnable;
    CullMode = 1;
    ShadeMode = Gouraud;
    AlphaBlendEnable = true;
    SrcBlend = SrcAlpha;
    DestBlend = One;
    AlphaTestEnable = true;
    AlphaRef = 1;
    AlphaFunc = GreaterEqual;
    Lighting = false;
    FogEnable = sFogEnable;
    VertexShader = compile vs_2_0 VertexShaderFunction2();
    PixelShader  = compile ps_2_0 PixelShaderFunction2();
  }  
} 

// Fallback
technique fallback
{
  pass P0
  {
    // Just draw normally
  }
}

